#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
#
#  Copyright 2002  The FreeRADIUS server project
#  Copyright 2002  Boian Jordanov <bjordanov@orbitel.bg>
#  Copyright 2012  Thomas Glanzmann <thomas@glanzmann.de>

use strict;
use Authen::Krb5::Simple;
use BerkeleyDB;
use Mail::Mailer;

use vars qw(%RAD_REQUEST %RAD_REPLY %RAD_CHECK);

use constant RLM_MODULE_REJECT => 0;   # immediately reject the request
use constant RLM_MODULE_FAIL => 1;     # module failed, don't reply
use constant RLM_MODULE_OK => 2;       # the module is OK, continue
use constant RLM_MODULE_HANDLED => 3;  # the module handled the request, so stop.
use constant RLM_MODULE_INVALID => 4;  # the module considers the request invalid.
use constant RLM_MODULE_USERLOCK => 5; # reject the request (user is locked out)
use constant RLM_MODULE_NOTFOUND => 6; # user not found
use constant RLM_MODULE_NOOP => 7;     # module succeeded without doing anything
use constant RLM_MODULE_UPDATED => 8;  # OK (pairs modified)
use constant RLM_MODULE_NUMCODES => 9; # How many return codes there are

my %fail;
my %locked;
my %tokens;
my $otp_lifetime = 600;
my $db_file = '/home/sithglan/work/smsotpd/users';
my $sender = 'thomas@glanzmann.de';
my $url = 'https://terminal.gmvl.de/';
my $strikes = 3;

my $db = new BerkeleyDB::Hash
        -Filename => $db_file,
        -Flags    => DB_RDONLY
        or die "Cannot open file $db_file: $! $BerkeleyDB::Error\n" ;

sub
send_otp
{
        my %args = @_;

        my $mailer = Mail::Mailer->new('sendmail');
        $mailer->open({
                        From => $sender,
                        To => "$args{to}\@sms.glanzmann.de",
                        Subject => $args{otp},
                        });
        $mailer->close();
}

sub
send_email
{
        my %args = @_;

        my $mailer = Mail::Mailer->new('sendmail');
        $mailer->open({
                        From => $sender,
                        To => $sender,
                        Subject => "SMSOTP: Account $args{user} locked"
                        });
        print $mailer <<EOF;
Hello,
the user account $args{user} was locked after $strikes failed logins.
In order to unlock the account go to $url
and login as user '$args{user}' with the unlock password '$args{password}'.

Cheers,
        rlm_perl
EOF
        $mailer->close();
}

sub
generate_otp
{
        my $user = lc($RAD_REQUEST{'User-Name'});

        my $value = undef;

        if ($db->db_get($user, $value) == 0) {
                $tokens{$user}->{id} = int(1 + rand(99999999));
                $tokens{$user}->{otp} = sprintf("%05d", int(1 + rand(99999)));
                $tokens{$user}->{time} = time;

                send_otp(to => $value, otp => $tokens{$user}->{otp});

                $RAD_REPLY{'Reply-Message'} = 'Enter SMS one time password';
                $RAD_REPLY{'State'} = $tokens{$user}->{id};
                $RAD_CHECK{'Response-Packet-Type'} = 'Access-Challenge'; 
                return RLM_MODULE_HANDLED;
        }

        return RLM_MODULE_REJECT;
}

sub
check_otp
{
        my $user = lc($RAD_REQUEST{'User-Name'});

        my $state = $RAD_REQUEST{'State'};
        $state =~ s/^0x//;
        $state =~ s/([a-fA-F0-9][a-fA-F0-9])/chr(hex($1))/eg;

        if (($tokens{$user}->{otp} eq $RAD_REQUEST{'User-Password'})
        && ($tokens{$user}->{id} eq $state)
        && ((time() - $tokens{$user}->{time}) < $otp_lifetime)) {
                delete($tokens{$user});
                return RLM_MODULE_OK;
        }

        delete($tokens{$user});
        return RLM_MODULE_REJECT;
}

sub
generate_password
{
        my $length = 8;
        my $possible = '#!"#$%&\'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~';

        my $password;

        my $length_possible = length($possible);

        while (length($password) < $length) {
                $password .= substr($possible, int(rand($length_possible)), 1);
        }

        return $password;
} 

sub authorize
{
        return RLM_MODULE_OK;
}

sub
authenticate
{
        my $user = lc($RAD_REQUEST{'User-Name'});

        my $value;

        if ($db->db_get($user, $value) != 0) {
                return RLM_MODULE_REJECT;
        }

        if (defined($locked{$user})) {
                if ($RAD_REQUEST{'User-Password'} eq $locked{$user}) {
                        delete($locked{$user});
                        return RLM_MODULE_REJECT;
                }
                return RLM_MODULE_USERLOCK;
        }

        if (defined($RAD_REQUEST{'State'})) {
                return check_otp();
        }

        if (Authen::Krb5::Simple->new()->authenticate($RAD_REQUEST{'User-Name'}, $RAD_REQUEST{'User-Password'})) {
                delete($fail{$user});
                return generate_otp();

        } else {
                $fail{$user}++;
                if ($strikes <= $fail{$user}) {
                        delete($fail{$user});
                        $locked{$user} = generate_password();
                        send_email(user => $user, password => $locked{$user});
                }
                return RLM_MODULE_REJECT;
        }
}

sub
preacct
{
        return RLM_MODULE_OK;
}

sub
accounting
{
        return RLM_MODULE_OK;
}

sub
checksimul
{
        return RLM_MODULE_OK;
}

sub
pre_proxy
{
        return RLM_MODULE_OK;
}

sub
post_proxy
{
        return RLM_MODULE_OK;
}

sub
post_auth
{
        return RLM_MODULE_OK;
}

sub
xlat
{
        # nothing todo
}

sub
detach
{
        # nothing todo
}
